;
;  Copyright (C) 2005  Roland Gruber
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
;  
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
;  
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;
;  This program uses an Opel TID to display various messages.
;
;  Changelog:
;  ----------
;
;  1.0 (24.03.2005): initial release
;


; Ports
#define		iSCL	PORTA,1		; SCL for I2C
#define		iSDA	PORTA,2		; SDA for I2C
#define		iSDAin	PORTA,3		; SDA in for I2C

#define		DIS	PORTB,1		; Display on
#define		SDA	PORTB,2		; SDA TID
#define		SCL	PORTB,3		; SCL TID
#define		MRQ	PORTB,4		; MRQ TID

#define		LED	PORTB,5		; LED

; Programm memory
buf		equ	0ch
count		equ	0dh
CounterA	EQU	0eh
CounterB	EQU	0fh
CounterC	EQU	10h
CounterD	EQU	11h
W1		EQU	12h
W2		EQU	13h
W3		EQU	14h
I2CBYTE		EQU	15h
AREG		EQU	16h
BREG		EQU	17h
CREG		EQU	18h
DREG		EQU	19h
TID1		EQU	1ah
TID2		EQU	2bh
TID3		EQU	2ch
TID4		EQU	2dh
TID5		EQU	2eh
TID6		EQU	2fh
TID7		EQU	20h
TID8		EQU	21h
TIDtemp		EQU	22h
TIDtextNr	EQU	23h

; Internal EEPROM
eepTID		EQU	01h	; address of TID message number


; Configuration
	list p=16f84,f=inhx8m
	Title "tid.asm"

	include p16f84.inc
	
	__CONFIG	_PWRTE_ON & _WDT_OFF & _XT_OSC

	org	00h
start
; Port B is output
	bsf     STATUS,RP0
	clrf    TRISB
	bcf     STATUS,RP0
	clrf	PORTB
; Port A is output, only iSDAin is input
	bsf     STATUS,RP0
	movlw	B'00001000'	; PortA all output excluding SDAin
	movwf	TRISA
	bcf     STATUS,RP0
	clrf	PORTA

; LED test
	bsf	LED
	call	delay_1_sec
	bcf	LED
	call	delay_1_sec
	bsf	LED

; init I2C of EEPROM
	call	i2c_reset

; start with TID text 0
	movlw	0x00
	movwf	TIDtextNr

; welcome message
	call	TIDon

	movlw	0x53
	call	addCharTID	; "S"

	movlw	0x45
	call	addCharTID	; "E"

	movlw	0x52
	call	addCharTID	; "R"

	movlw	0x56
	call	addCharTID	; "V"

	movlw	0x55
	call	addCharTID	; "U"

	movlw	0x53
	call	addCharTID	; "S"

	call	TIDoff

	call	delay_300_sec
	call	show_messages

myloop
	bsf	LED
	call	wait01s
	call	wait01s
	call	wait01s
	call	wait01s
	call	wait01s
	bcf	LED
	call	wait01s
	call	wait01s
	call	wait01s
	call	wait01s
	call	wait01s
	goto myloop

;-------------------
; Message displaying
;-------------------

; show messages from EEPROM
show_messages
	call nextAddress
	call TIDon		; turn TID on
	call i2c_start		; I2C START
	movlw	B'10100000'	; Control
	call	i2c_tx
	movf	TIDtextNr,0	; Address 0
	call	i2c_tx
	movlw	0x00		; Address 1
	call	i2c_tx
	call	i2c_start	; 2. Start
	movlw	B'10100001'	; 2. Control
	call	i2c_tx
show_messages_next
	call	i2c_rxack
	movf	buf,1		; test if zero
	btfsc	STATUS,Z	; 0?
	goto	show_messages_end
	call	addCharTID	; else display char
	goto	show_messages_next
show_messages_end
	call	i2c_rx
	call	i2c_stop	; I2C STOP
	call	TIDoff		; turn TID off
	call	delay_300_sec
	goto	show_messages

; get I2C memory address for next text
nextAddress
	; read message number from internal EEPROM
	movlw	eepTID		; address of TID message number
	movwf	EEADR
	bsf	STATUS,RP0	; Bank 1
	bsf	EECON1,RD	; read
	bcf	STATUS,RP0	; Bank 1
	movf	EEDATA,0	; read from EEPROM
	movwf	TIDtextNr	; to TIDtextNr
	incf	TIDtextNr,1	; show next message
	; write new message number to internal EEPROM
	movf	TIDtextNr,0	; write data
	movwf	EEDATA
	movlw	eepTID		; address of TID message number
	movwf	EEADR
	bsf	STATUS,RP0	; Bank 1
	bcf	INTCON,GIE	; disable interupts
	bsf	EECON1,WREN	; write enable
	movlw	0x55
	movwf	EECON2		; write 55h
	movlw	0xaa
	movwf	EECON2		; write AAh
	bsf	EECON1,WR	; begin write
nextAddress_write
	btfsc	EECON1, WR
	goto	nextAddress_write
	bcf	EECON1,WREN	; write disable
	bsf	INTCON,GIE	; reenable interupts
	bcf	STATUS,RP0	; Bank 0
	return
	
;----------------
; I2C subroutines
;----------------

i2c_reset
	bsf	iSDA
	bsf	iSCL
	nop
	movlw	9
	movwf	buf
i2c_reset1
	bcf	iSCL
	nop
	nop
	nop
	bsf	iSCL
	nop
	nop
	nop
	decfsz	buf, f
	goto	i2c_reset1
	nop
	bsf	iSCL
	return

i2c_start
	bsf	iSCL
	bsf	iSDA
	nop
	bcf	iSDA
	nop
	bcf	iSCL
	return

i2c_stop
	nop
	nop
	bsf	iSCL
	nop
	bsf	iSDA
	return

; I2C: send w
i2c_tx
	call	WrI2cW
	bsf	iSDA
	; ACK
	bsf	iSCL
	nop
	nop
	nop
	bcf	iSCL
	bcf	iSDA
	return 

; I2C: read w with ACK
i2c_rxack
	call	RdI2cW
	bcf	iSDA
	nop
	nop
	nop
	nop
	; ACK
	bsf	iSCL
	nop
	nop
	bcf	iSCL
	bcf	iSDA
	return

; I2C: read w without ACK
i2c_rx
	call	RdI2cW
	nop
	nop
	bsf	iSDA
	nop
	bsf	iSCL
	nop
	nop
	bcf	iSCL
	bcf	iSDA
	return

; I2C: send w to EEPROM
WrI2cW
	movwf	buf
	movlw	8
	movwf	count
WrI2cW1

	bcf	iSDA
	rlf	buf,f
	btfsc	STATUS,C	; 0?
	bsf	iSDA		; no, 1
	nop
	bsf	iSCL
	nop
	nop
	bcf	iSCL
	decfsz	count,f
	goto	WrI2cW1
	return
			


; I2C: read from EEPROM to w
RdI2cW
	clrf	buf
	movlw	8
	movwf	count
	bsf	iSDA
RdI2cW1
	nop
	clrc
	btfsc	iSDAin
	setc
	rlf	buf,f

	bsf	iSCL
	nop
	nop
	bcf	iSCL
	decfsz	count,f
	goto	RdI2cW1
	movfw	buf
	return


;----------------
; TID subroutines
;----------------

; TID scrolling
; shift text one char to right and add new char from W
; then send to TID
addCharTID
	movwf	TIDtemp
	movf	TID2,0
	movwf	TID1
	movf	TID3,0
	movwf	TID2
	movf	TID4,0
	movwf	TID3
	movf	TID5,0
	movwf	TID4
	movf	TID6,0
	movwf	TID5
	movf	TID7,0
	movwf	TID6
	movf	TID8,0
	movwf	TID7
	movf	TIDtemp,0
	movwf	TID8
	call	delay_0.2_sec
	call	sendTIDText
	return

; Sends 8 chars to TID
sendTIDText
	call	startTID
	movf	TID1,0
	call	sendTID
	movf	TID2,0
	call	sendTID
	movf	TID3,0
	call	sendTID
	movf	TID4,0
	call	sendTID
	movf	TID5,0
	call	sendTID
	movf	TID6,0
	call	sendTID
	movf	TID7,0
	call	sendTID
	movf	TID8,0
	call	sendTID
	call	stopTID
	return

; Display on; init TID
TIDon
	; init TID text (all spaces)
	movlw 	0x20
	movwf 	TID1
	movwf 	TID2
	movwf 	TID3
	movwf 	TID4
	movwf 	TID5
	movwf 	TID6
	movwf 	TID7
	movwf 	TID8
	bsf	DIS		; Display on
	call	wait01s
	; send empty text 10 times
	movlw	0x05
	movwf	DREG
TIDon1	call	sendTIDText
	call	wait01s
	decfsz	DREG
	goto	TIDon1
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x20
	call	addCharTID	; " "
	return

; Display off
; scroll out text
TIDoff
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x2a
	call	addCharTID	; "*"
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	movlw	0x20
	call	addCharTID	; " "
	bcf	DIS		; Display off (show date)
	return


; Start condition for TID
startTID
	bcf	MRQ
	bcf	SDA		; All high
	bcf	SCL	;
	; communication begins
	bsf	MRQ	;
	call	wait1ms		; MRQ neg. imp.
	bcf	MRQ	;

	call	wait1ms
	; I2C-start
	bsf	SDA		; START:
	call	wait1ms		; SCL=1, SDA=falling
	bsf	SCL		;

	; address
	movlw	0x4a		; Addres
	call	sendTID

	bsf	MRQ		; After the addres MRQ=0

	clrw
	call	sendTID		; All lamp off
	clrw
	call	sendTID
	return

; Stop condition for TID
stopTID
	call	wait1ms
	bcf	MRQ		; MRQ=1

	bsf	SDA
	call	wait1ms
	bcf	SCL		; Stop:
	call	wait1ms
	bcf	SDA		; SCL=1, SDA=raising
	call	wait1ms

	bcf	MRQ
	bcf	SDA		; End: all high
	bcf	SCL
	return

; Send W to TID
; Calculate the parity
sendTID
	movwf	BREG		; Send byte (W)
	movwf	I2CBYTE		; Send byte (W)
	clrf	CREG		; The number of 1
	movlw	0x07
	movwf	AREG		; cycle value

send0	rrf	BREG		; rotate the byte,
	btfsc	STATUS,C	; and count the number of 1
	incf	CREG		; (but only the low 7 bit)
	decfsz	AREG,1
	goto	send0

	rlf	I2CBYTE,1	; Rotate, and behind the parity
	bcf	I2CBYTE,0
	btfss	CREG,0		; if CREG odd, then clear parity
	bsf	I2CBYTE,0	;

	; send byte
	movlw	8		;
	movwf	CREG		; Cycle 8x

send1
	bsf	SCL		; Clock=0
	call	wait025ms

	btfsc	I2CBYTE,7	; byte <7> bit 0 or 1
	goto	send3
	bsf	SDA		; 0
	goto	send4
send3
	bcf	SDA		; 1
send4
	call	wait025ms
	bcf	SCL		; Clock=1, valid data
	call	wait025ms

	rlf	I2CBYTE,1	; rotate

	decfsz	CREG,1		; Cycle
	goto	send1

	bsf	SCL		; Clock = 0
	call	wait025ms

	; ACK
	bcf	SDA		; Output=1
	call	wait025ms
	bcf	SCL		; Clock=1=> ACK
	
	call	wait025ms	; ignore ACK

	bsf	SCL		; Clock=0
	call	wait025ms
	bsf	SDA		; data=0
	return


;-------------------
; Timing subroutines
;-------------------

; Delay code generated by PiKloops (Mo 2005-Mr-14 20:56:41)
; Time Delay = 0.00025000s  with  Osc = 4.00000000MHz
wait025ms
	movlw 	D'83'
	movwf 	CounterA
delay_0.00025_sec_loop
	decfsz 	CounterA,1
 	goto 	delay_0.00025_sec_loop 
	return

; Delay code generated by PiKloops (Mo 2005-Mr-14 20:55:35)
; Time Delay = 0.00100000s  with  Osc = 4.00000000MHz
wait1ms
	movlw 	D'2'
	movwf 	CounterB
	movlw 	D'75'
	movwf 	CounterA
delay_0.001_sec_loop
	decfsz 	CounterA,1
 	goto 	delay_0.001_sec_loop 
	decfsz 	CounterB,1
 	goto 	delay_0.001_sec_loop 
	return

; Delay code generated by PiKloops (Mo 2005-Mr-14 20:54:18)
; Time Delay = 0.09999500s  with  Osc = 4.00000000MHz
wait01s
	movlw 	D'130'
	movwf 	CounterB
	movlw 	D'220'
	movwf 	CounterA
delay_0.1_sec_loop
	decfsz 	CounterA,1
 	goto 	delay_0.1_sec_loop 
	decfsz 	CounterB,1
 	goto 	delay_0.1_sec_loop 
	return

; Delay code generated by PiKloops (Sa 2005-Mr-19 14:17:32)
; Time Delay = 0.19999300s  with  Osc = 4.00000000MHz
delay_0.2_sec
	movlw 	D'2'
	movwf 	CounterC
	movlw 	D'5'
	movwf 	CounterB
	movlw 	D'183'
	movwf 	CounterA
delay_0.2_sec_loop
	decfsz 	CounterA,1
 	goto 	delay_0.2_sec_loop 
	decfsz 	CounterB,1
 	goto 	delay_0.2_sec_loop 
	decfsz 	CounterC,1
 	goto 	delay_0.2_sec_loop 
	return

; Delay code generated by PiKloops (Mo 2005-Feb-28 20:42:10)
; Time Delay = 0.99999800s  with  Osc = 4.00000000MHz
delay_1_sec
	movlw 	D'6'
	movwf 	CounterC
	movlw 	D'24'
	movwf 	CounterB
	movlw 	D'168'
	movwf 	CounterA
delay_1_sec_loop
	decfsz 	CounterA,1
	goto 	delay_1_sec_loop 
	decfsz 	CounterB,1
	goto 	delay_1_sec_loop 
	decfsz 	CounterC,1
	goto 	delay_1_sec_loop 
	return

; Delay code generated by PiKloops (Mo 2005-Mr-14 21:15:22)
; Time Delay = 9.99999800s  with  Osc = 4.00000000MHz
delay_10_sec
	movlw 	D'51'
	movwf 	CounterC
	movlw 	D'237'
	movwf 	CounterB
	movlw 	D'173'
	movwf 	CounterA
delay_10_sec_loop
	decfsz 	CounterA,1
	goto 	delay_10_sec_loop 
	decfsz 	CounterB,1
	goto 	delay_10_sec_loop 
	decfsz 	CounterC,1
	goto 	delay_10_sec_loop 
	return

; Delay code generated by PiKloops (Mi 2005-Mr-16 19:05:14)
; Time Delay = 300.00000000s  with  Osc = 4.00000000MHz
delay_300_sec
	movlw 	D'6'
	movwf 	CounterD
	movlw 	D'253'
	movwf 	CounterC
	movlw 	D'216'
	movwf 	CounterB
	movlw 	D'109'
	movwf 	CounterA
delay_300_sec_loop
	decfsz 	CounterA,1
 	goto 	delay_300_sec_loop 
	decfsz 	CounterB,1
 	goto 	delay_300_sec_loop 
	decfsz 	CounterC,1
 	goto 	delay_300_sec_loop 
	decfsz 	CounterD,1
 	goto 	delay_300_sec_loop 
	return

end